# -*- coding: utf-8 -*-
"""
This sub-package defines all the machinery for describing and running
transfomations. A transformation consists in a series of action steps, each
one taking an input object (e.g. an XML tree), performing some transformations
and returning an output object (e.g. the same but modified XML tree).
"""
